	/* Calcule sgn(det[ [a,b,c], [d,e,f], [g,h,i] ]) */
	/* de maniere exacte, avec des doubles representant des entiers */
	/* compris dans [-2^53;2^53]. */

	/* Sylvain Pion,   le 29/6/95 */

#include <stdio.h>
#include <math.h>

#define FILTRE0 1024.0*1024.0*1024.0*1024.0*1024.0	/* 2^50  */
#define FILTRE3 FILTRE0*256.0                           /* 2^58  */
#define FILTRE2 FILTRE3*1024.0*1024.0*128.0		/* 2^(58+27) */
#define FILTRE1 FILTRE2*1024.0*1024.0*128.0		/* 2^(58+54) */

#define DEC1 3.0*256.0*FILTRE0*1024.0*1024.0		/* 3.2^78 */
#define DEC2 DEC1*1024.0*1024.0*128.0			/* 3.2^(78+27) */
#define DEC3 DEC2*1024.0*1024.0*128.0			/* 3.2^(78+54) */
#define DEC4 DEC3*1024.0*1024.0*128.0			/* 3.2^(78+54+27) */

#define deco(a,a1,a2,t) a2 = a - (a1 = (a + t) - t)
#define decomp1(x,y,z) deco(x,y,z,DEC1);
#define decomp2(x,y,z) deco(x,y,z,DEC2);
#define decomp3(x,y,z) deco(x,y,z,DEC3);
#define decomp4(x,y,z) deco(x,y,z,DEC4);

	/* On peut utiliser des pointeurs, il n'y a pas modification */
	/* des valeurs de a, b, c, d, e, f, g, h, i. */


int signe3(     double a, double b, double c,
                double d, double e, double f,
                double g, double h, double i)
{
	double p;

	p =	a * (e * i - f * h) +
		b * (f * g - i * d) +
		c * (d * h - g * e);

	/* Est-ce que ca vaut le coup en moyenne ??? */

	if (fabs(p) >= FILTRE1)	return( (p>0) ? 1 : -1 );
	else {
		double	a1,a2, b1,b2, c1,c2,

                        z001, z011, z021, z031,
                        z101, z111, z121, z131,
                        z201, z211, z221, z231,

                        z002, z012, z022, z032,
                        z102, z112, z122, z132,
                        z202, z212, z222, z232,

			p011, p021, p031, p041, p051, p061, p071, p081,
			p111, p121, p131, p141, p151, p161, p171, p181,
			p211, p221, p231, p241, p251, p261, p271,

			p012, p022, p032, p042, p052, p062, p072, p082,
			p112, p122, p132, p142, p152, p162, p172, p182,
			p212, p222, p232, p242, p252, p262, p272;

		{
		double	z00, z01, z02, z03,
                        z10, z11, z12, z13,
                        z20, z21, z22, z23;
		{
		double	d1,d2, e1,e2, f1,f2,
                        g1,g2, h1,h2, i1,i2;

		decomp1(a,a1,a2); decomp1(b,b1,b2); decomp1(c,c1,c2);
		decomp1(d,d1,d2); decomp1(e,e1,e2); decomp1(f,f1,f2);
                decomp1(g,g1,g2); decomp1(h,h1,h2); decomp1(i,i1,i2);

		z00 = e1*i1 - f1*h1;
                z01 = e1*i2 - f1*h2;
                z02 = e2*i1 - f2*h1;
                z03 = e2*i2 - f2*h2;

                z10 = f1*g1 - i1*d1;
                z11 = f1*g2 - i1*d2;
                z12 = f2*g1 - i2*d1;
                z13 = f2*g2 - i2*d2;
                
                z20 = d1*h1 - e1*g1;
                z21 = d1*h2 - e1*g2;
                z22 = d2*h1 - e2*g1;
                z23 = d2*h2 - e2*g2;

		};

		decomp1(z03,z031,z032);
                decomp1(z13,z131,z132);
                decomp1(z23,z231,z232);

                decomp2(z02,z021,z022);
                decomp2(z12,z121,z122);
                decomp2(z22,z221,z222);

                decomp2(z01,z011,z012);
                decomp2(z11,z111,z112);
                decomp2(z21,z211,z212);

                decomp3(z00,z001,z002);
                decomp3(z10,z101,z102);
                decomp3(z20,z201,z202);

		};

		p = a1*z001 + b1*z101 + c1*z201;

		{
		double p01, p02, p11, p12, p21, p22;

		p01 = a1*z002 + a2*z001;
		p02 = a1*(z021 + z011);
                p11 = b1*z102 + b2*z101;
                p12 = b1*(z121 + z111);
                p21 = c1*z202 + c2*z201;
                p22 = c1*(z221 + z211);

		decomp4(p01,p011,p012);
                decomp4(p02,p021,p022);
                decomp4(p11,p111,p112);
                decomp4(p12,p121,p122);
                decomp4(p21,p211,p212);
                decomp4(p22,p221,p222);

		};

		p += (p011 + p021 + p111 + p121 + p211 + p221);

		{
		double p03, p04, p05, p13, p14, p15, p23, p24, p25;

		p03 = a1*z031 + a2*z021;
		p04 = a1*(z012 + z022);
		p05 = a2*(z011 + z002);
                p13 = b1*z131 + b2*z121;
                p14 = b1*(z112 + z122);
                p15 = b2*(z111 + z102);
                p23 = c1*z231 + c2*z221;
                p24 = c1*(z212 + z222);
                p25 = c2*(z211 + z202);

		decomp3(p03,p031,p032);
                decomp3(p04,p041,p042);
                decomp3(p05,p051,p052);
                decomp3(p13,p131,p132);
                decomp3(p14,p141,p142);
                decomp3(p15,p151,p152);
                decomp3(p23,p231,p232);
                decomp3(p24,p241,p242);
                decomp3(p25,p251,p252);

		};

		p +=	( p012 + p022 + p112 + p122 + p212 + p222
			+ p031 + p041 + p051 + p131 + p141
			+ p151 + p231 + p241 + p251);

                if (fabs(p) >= FILTRE2) return ((p>0)? 1 : -1);

		{
		double p06, p07, p16, p17, p26, p27;

		p06 = a1*z032 + a2*z031;
		p07 = a2*(z022 + z012);
                p16 = b1*z132 + b2*z131;
                p17 = b2*(z122 + z112);
                p26 = c1*z232 + c2*z231;
                p27 = c2*(z222 + z212);

                decomp2(p06,p061,p062);
		decomp2(p07,p071,p072);
                decomp2(p16,p161,p162);         
                decomp2(p17,p171,p172);
                decomp2(p26,p261,p262);         
                decomp2(p27,p271,p272);

		};

		p += (  p032 + p042 + p052 + p132 + p142 +
			p152 + p232 + p242 + p252 +
			p061 + p071 + p161 + p171 + p261 + p271);

                if (fabs(p) >= FILTRE3) return ((p>0)? 1 : -1);

		{
		double p08, p18;

                p08 = a2*z032;
                p18 = b2*z132 + c2*z232;

		decomp1(p08,p081,p082);
                decomp1(p18,p181,p182);

		};

		p += (p062 + p072 + p162 + p172 + p262 + p272 + p081 + p181);

	/* Filtre superflu */

		p += (p082 + p182);

		if (p>0) return 1; else return( (p<0) ? -1 : 0 );
	};
}
