#include <stream.h>
#include <math.h>
#include "prims.h"
#include "ABDPY.h"
#include "det2x2.h"
#include "det3x3.h"
#include "arith.h"
#include "AR2x2.h"
#include "AR3x3.h"

extern "C" int getpid();
extern "C" long clock();

void exact_out(ostream& s, double a)
{
 s.precision(20);
 s<<a;
 s.precision(6);
}


// random value generation
// return in a double a signed integer with i bits
double rnd(int i)
{ if (i < 25 )
	return  (((int) random()%2)?1:-1) *
	((double) ((int) random()% ((int)exp2(i)) ));
  else
	return  (((int) random()%2)?1:-1) *
	(((double) ((int) random()% ((int)exp2(i-25)) )) * exp2(25) +
	((double) ((int) random()% ((int)exp2(25)) )));
}


// Few 2x2 determinant sign computation
int quadruple_det2x2(double a, double b, double c, double d)
{
long double DET=(long double)a*(long double)d-(long double)b*(long double)c;
if (DET>0.0) return 1;
if (DET<0.0) return -1;
return 0;
}

int double_det2x2(double a, double b, double c, double d)
{ return ( ((a*d)<(b*c)) ? -1 : 1) ; }

int double_det3x3(double a, double b, double c,
		   double d, double e, double f, double g, double h, double i)
{ if ( a*(e*i-f*h) + b*(f*g-d*i) + c*(d*h-e*g) < 0 ) return -1 ;
  else return 1; }






void generate2(double& a,double& b,double& c,double& d, char type)
{
  double cxx,cyy;
  switch (type) {
  case 'r' :               // random
    a = rnd(53);
    b = rnd(53);
	c = rnd(53);
    d = rnd(53);
	break;

  case 'p' :              // x+y=0
    a = rnd(53);
    b = -a;
	c = rnd(53);
    d = -c;
    break;

  case 'q' :              // aU bV
    cxx = rnd(26);
    cyy = rnd(26);
    a =   rnd(27);
    c =   rnd(27);
    b = cxx * a;
    d = cxx * c;
	a *= cyy;
	c *= cyy;
    break;

  case 'f' :              // y = floor(u*x)      u in [0,1]
    cxx = rnd(52);
    a = rnd(53);
    b = floor(a*cxx/exp2(52));
	c = rnd(53);
    d = floor(c*cxx/exp2(52));
	break;

  case 'c' :              // all entries equal
    a = b = c = d = rnd(53);
	break;
}}

void generate3(double& a,double& b,double& c,double& d,double& e,
			   double& f,double& g,double& h,double& i, char type)
{
  double cxx,cyy;
   switch (type) {
  case 'r' :              // random
    a = rnd(51);
    b = rnd(51);
	c = rnd(51);
    d = rnd(51);
    e = rnd(51);
    f = rnd(51);
    g = rnd(51);
	h = rnd(51);
    i = rnd(51);
	break;

  case 'q' :            // aU bV cU+dV
    cxx = rnd(25);
    cyy = rnd(25);
    a =   rnd(25);
    b =   rnd(25);
    c =   cxx * a + cyy * b;
    d =   rnd(26);
    e =   rnd(26);
    f =   cxx * d + cyy * e;
    g =   rnd(25);
    h =   rnd(25);
    i =   cxx * g + cyy * h;
    cxx = rnd(25);
    cyy = rnd(25);
	a *= cxx;
	d *= cxx;
	g *= cxx;
	b *= cyy;
	e *= cyy;
	h *= cyy;
    break;

  case 'p' :              // x+y+z=0
    a = rnd(50);
    b = rnd(50);
	c = -a -b;
    d = rnd(50);
    e = rnd(50);
    f = -d -e;
    g = rnd(50);
	h = rnd(50);
    i = -g -h;
    break;

  case 'f' :              // z = floor( t*x + u*y )     t,u in [0,1]
    cxx = rnd(51);
	cyy = rnd(51);
    a = rnd(51);
    b = rnd(51);
	c = floor((-a*cxx-b*cyy)/exp2(52));
    d = rnd(51);
    e = rnd(51);
    f = floor((-d*cxx-e*cyy)/exp2(52));
	g = rnd(51);
	h = rnd(51);
	i = floor((-g*cxx-h*cyy)/exp2(52));
	break;

  case 'c' :             // all entries equal
    a = b = c = d = e = f = g = h = i = rnd(51);
	break;
}}

void perturbation2(double& a,double& b,double& c,double& d)
{ a += rnd(2); b += rnd(2); c += rnd(2); d += rnd(2);}

void perturbation3(double& a,double& b,double& c,double& d,double& e,
			   double& f,double& g,double& h,double& i)
{ a += rnd(2); b += rnd(2); c += rnd(2); d += rnd(2); e += rnd(2); f += rnd(2);
	g += rnd(2); h += rnd(2); i += rnd(2); }

void transpose2(double& a,double& b,double& c,double& d)
{ double swap; (void) a; (void) d; swap = b; b = c; c = swap; }

void transpose3(double& a,double& b,double& c,double& d,double& e,
			   double& f,double& g,double& h,double& i)
{ double swap; swap = b; b = d; d = swap; swap = c; c = g; g = swap;
	swap = f; f = h; h = swap; (void) a; (void) e; (void) i; }





void interactive()
{
	int dim;
	double a,b,c,d,e,f,g,h,i;

	cout << " dim ? ";
	cin >> dim;
	while (1) if (dim == 2) {
	   cout << "a = "; cin >> a; 
	   cout << "b = "; cin >> b;
	   cout << "c = "; cin >> c;
	   cout << "d = "; cin >> d;
	   cout << endl;
	   cout << " | " ; exact_out(cout,a);
       cout << " " ; exact_out(cout,c);
       cout << " |" << endl;
	   cout << " | " ; exact_out(cout,b);
       cout << " " ; exact_out(cout,d);
       cout << " |" << endl;
	 cout << "double :" << double_det2x2(a,b,c,d) << endl;
	 cout << "quadruple :" << quadruple_det2x2(a,b,c,d) << endl;
	 cout << "LN :" << LN_det2x2(a,b,c,d) << endl;
	 cout << "ABDPY not lazy:" << ABDPY__not_lazy_det2x2(a,b,c,d) << endl;
	 cout << "ABDPY lazy:" << ABDPY_det2x2(a,b,c,d) << endl;
	 cout << "ABDPY secure:" << ABDPY_secure_det2x2(a,b,c,d) <<  endl;
	 cout << "det2x2:" << signe(a,b,c,d) << endl;
	 cout << "AR_det2: " << AR_det2 (a,b,c,d) << endl;
	} else if (dim == 3) {
	   cout << "a = "; cin >> a;
	   cout << "b = "; cin >> b;
	   cout << "c = "; cin >> c;
	   cout << "d = "; cin >> d;
	   cout << "e = "; cin >> e;
	   cout << "f = "; cin >> f;
	   cout << "g = "; cin >> g;
	   cout << "h = "; cin >> h;
	   cout << "i = "; cin >> i;
	   cout << " | " ; exact_out(cout,a);
       cout << " " ; exact_out(cout,d);
       cout << " " ; exact_out(cout,g);
       cout << " |" << endl;
	   cout << " | " ; exact_out(cout,b);
       cout << " " ; exact_out(cout,e);
       cout << " " ; exact_out(cout,h);
       cout << " |" << endl;
	   cout << " | " ; exact_out(cout,c);
       cout << " " ; exact_out(cout,f);
       cout << " " ; exact_out(cout,i);
       cout << " |" << endl;
	 cout << "double :" << double_det3x3(a,b,c,d,e,f,g,h,i) << endl;
	 cout << "LN :" << LN_det3x3(a,b,c,d,e,f,g,h,i) << endl;
	 cout << "ABDPY not lazy:"<<ABDPY__not_lazy_det3x3(a,b,c,d,e,f,g,h,i)<<endl;
	 cout << "ABDPY lazy:" << ABDPY_det3x3(a,b,c,d,e,f,g,h,i) << endl;
	 cout << "ABDPY secure:" << ABDPY_secure_det3x3(a,b,c,d,e,f,g,h,i) << endl;
         cout << "det3x3:" << signe3(a,b,c,d,e,f,g,h,i) << endl;
         cout << "AR_det3:" << AR_det3(a,b,c,d,e,f,g,h,i) << endl;
	}
}




main(int argc, char **argv)
{ double a,b,c,d,e,f,g,h,i;
  int sign;
  int n,N=100;
  //int n,N=80000;
  int input, method, trial, j;
  double time[10];
  long t1,t2;
  int (*FF2)(double, double, double, double);
  int (*FF3)(double, double, double, double, double, double, double, double, double);

 srandom(n=getpid());
 cerr <<"init random with pid ="<<n<<endl;

 if (--argc)
 {
	++argv;
	if (    (**argv>='1') && (**argv<='9') ) n  = atoi(*argv);
	else interactive();
 }
 else n = 10;

  if (!argc){
  cout <<"\\documentclass{article}" << endl;
  cout <<"\\begin{document}" << endl;
  cout <<"\\subsection*{Running time in $\\mu$s.}" << endl;
  cout <<"\\begin{tabular}{|l||r|r|r|r|r|r|r|r|}\\hline" << endl;
  cout <<"input&double&quadruple&LN&not-lazy&lazy&secure&det2x2&AR2x2\\\\\\hline" << 
endl;
  }

 for ( input = 0 ; input < 11 ; input++ )
 {
	switch (input) {
	  case  0: cout<< "random"; break;
	  case  1: cout<< "$x=-y$"; break;
	  case  2: cout<< "$x=-y+\\varepsilon$"; break;
	  case  3: cout<< "$x=-y^t$"; break;
	  case  4: cout<< "$x=-y+\\varepsilon^t$"; break;
	  case  5: cout<< "$kU,lU$"; break;
	  case  6: cout<< "$kU,lU+\\varepsilon$"; break;
	  case  7: cout<< "$U,\\lfloor\\alpha U\\rfloor$"; break;
	  case  8: cout<< "$U,\\lfloor\\alpha U\\rfloor^t$"; break;
	  case  9: cout<< "$=$"; break;
	  case 10: cout<< "$=+\\varepsilon$"; break;
	}

	for ( method = 0 ; method < 8 ; method++)  time[method] = 0.0;
	for (trial=0; trial < n ; trial++){
	  switch (input) {
	  	case  0: generate2(a,b,c,d,'r'); break;
	  	case  1: generate2(a,b,c,d,'p'); break;
	  	case  2: generate2(a,b,c,d,'p'); perturbation2(a,b,c,d); break;
	  	case  3: generate2(a,b,c,d,'p'); transpose2(a,b,c,d); break;
	  	case  4: generate2(a,b,c,d,'p'); perturbation2(a,b,c,d);transpose2(a,b,c,d); break;
	  	case  5: generate2(a,b,c,d,'q'); break;
	  	case  6: generate2(a,b,c,d,'q'); perturbation2(a,b,c,d); break;
	  	case  7: generate2(a,b,c,d,'f'); break;
	  	case  8: generate2(a,b,c,d,'f'); transpose2(a,b,c,d); break;
	  	case  9: generate2(a,b,c,d,'c'); break;
	  	case 10: generate2(a,b,c,d,'c'); perturbation2(a,b,c,d); break;
	  }
	  sign =  ABDPY_det2x2(a,b,c,d);
	  for ( method = 0 ; method < 8 ; method++) {
		switch (method) {
          case 0 : FF2 = double_det2x2; break;
          case 1 : FF2 = quadruple_det2x2; break;
	  case 2 : FF2 = LN_det2x2; break;
          case 3 : FF2 = ABDPY__not_lazy_det2x2; break;
          case 4 : FF2 = ABDPY_det2x2; break;
          case 5 : FF2 = ABDPY_secure_det2x2; break;
	  case 6 : FF2 = signe; break;
	  case 7 : FF2 = AR_det2; break;
		}
		if ( method) if (sign != FF2(a,b,c,d) )
		{ cerr << "ERROR det2x2- send the following lines to "
                         << "spion@sophia.inria.fr"
                         <<input<<" "<<trial<< " double "
                         << double_det2x2(a,b,c,d) << ", quadruple "
                         << quadruple_det2x2(a,b,c,d) << ", LN "
		         << LN_det2x2(a,b,c,d) << ", ABDPY"
                         << ABDPY__not_lazy_det2x2(a,b,c,d) << ", ABDPY"
                         << ABDPY_det2x2(a,b,c,d) << ", ABDPY"
                         << ABDPY_secure_det2x2(a,b,c,d) <<  ", det2x2"
                         << signe(a,b,c,d) <<  ", AR2x2"
			 << AR_det2(a,b,c,d) << endl;
          cerr<<"a= ";exact_out(cerr,a); cerr<<endl;
          cerr<<"b= ";exact_out(cerr,b); cerr<<endl;
          cerr<<"c= ";exact_out(cerr,c); cerr<<endl;
          cerr<<"d= ";exact_out(cerr,d); cerr<<endl;
	  cerr<<endl<<"send the above lines to spion@sophia.inria.fr"<<endl;
		}
		if (!argc)
		{	
			t1 = clock();
			for (j=0; j<N; j++)
							(void) FF2(a,b,c,d);
			t2 = clock()-t1;
		}
		time[method] += t2 / (double) N;
	  }// end for method
	} //end for trial
	if(!argc)for ( method = 0 ; method < 8 ; method++)cout<<"&"<<time[method]/n;
	if(!argc)cout << "\\\\\\hline";
    cout << endl;
  }  // end for input

  if (!argc){
  cout <<"\\end{tabular}" << endl;
  cout << endl;
  cout <<"\\subsection*{Running time in $\\mu$s.}" << endl;
  cout <<"\\begin{tabular}{|l||r|r|r|r|r|r|r|}\\hline" << endl;
  cout <<"input&double&LN&not-lazy&lazy&secure&det3x3\\\\\\hline" << endl;
  }

 for ( input = 0 ; input < 11 ; input++ )
 {
        switch (input) {
          case  0: cout<< "random"; break;
          case  1: cout<< "$x+y+z=0$"; break;
          case  2: cout<< "$x+y+z=0+\\varepsilon$"; break;
          case  3: cout<< "$x+y+z=0^t$"; break;
          case  4: cout<< "$x+y+z=0+\\varepsilon^t$"; break;
          case  5: cout<< "$kU,lV,mU+nV$"; break;
          case  6: cout<< "$kU,lV,mU+nV+\\varepsilon$"; break;
          case  7: cout<< "$U,V,\\lfloor\\alpha U+\\beta V\\rfloor$"; break;
          case  8: cout<< "$U,V,\\lfloor\\alpha U+\\beta V\\rfloor^t$"; break;
          case  9: cout<< "$=$"; break;
          case 10: cout<< "$=+\\varepsilon$"; break;
        }
        for ( method = 0 ; method < 8 ; method++)  time[method] = 0.0;
        for (trial=0; trial < n ; trial++){
                switch (input) {
                case  0: generate3(a,b,c,d,e,f,g,h,i,'r'); break;
                case  1: generate3(a,b,c,d,e,f,g,h,i,'p'); break;
                case  2: generate3(a,b,c,d,e,f,g,h,i,'p'); perturbation3(a,b,c,d,e,f,g,h,i); break;
                case  3: generate3(a,b,c,d,e,f,g,h,i,'p'); transpose3(a,b,c,d,e,f,g,h,i); break;
                case  4: generate3(a,b,c,d,e,f,g,h,i,'p'); perturbation3(a,b,c,d,e,f,g,h,i);transpose3(a,b,c,d,e,f,g,h,i);  break;
                case  5: generate3(a,b,c,d,e,f,g,h,i,'q'); break;
                case  6: generate3(a,b,c,d,e,f,g,h,i,'q'); perturbation3(a,b,c,d,e,f,g,h,i); break;
                case  7: generate3(a,b,c,d,e,f,g,h,i,'f'); break;
                case  8: generate3(a,b,c,d,e,f,g,h,i,'f'); transpose3(a,b,c,d,e,f,g,h,i); break;
                case  9: generate3(a,b,c,d,e,f,g,h,i,'c'); break;
                case 10: generate3(a,b,c,d,e,f,g,h,i,'c'); perturbation3(a,b,c,d,e,f,g,h,i); break;
                }
                sign =  ABDPY_det3x3(a,b,c,d,e,f,g,h,i);
          for ( method = 0 ; method < 7 ; method++) {
                switch (method) {
          case 0 : FF3 = double_det3x3; break;
          case 1 : FF3 = LN_det3x3; break;
          case 2 : FF3 = ABDPY__not_lazy_det3x3; break;
          case 3 : FF3 = ABDPY_det3x3; break;
          case 4 : FF3 = ABDPY_secure_det3x3; break;
	  case 5 : FF3 = signe3; break;
	  case 6 : FF3 = AR_det3; break;
                  }
                if ( method) if (sign != FF3(a,b,c,d,e,f,g,h,i) )
                { cerr << "ERROR det3x3- send the following lines to "
                         << "spion@sophia.inria.fr"
                         << input<<" "<<trial<< " double "
                         << double_det3x3(a,b,c,d,e,f,g,h,i) << ", LN "
                         << LN_det3x3(a,b,c,d,e,f,g,h,i) << ", ABDPY "
                         << ABDPY__not_lazy_det3x3(a,b,c,d,e,f,g,h,i)<<", ABDPY "
                         << ABDPY_det3x3(a,b,c,d,e,f,g,h,i) << ", ABDPY "
                         << ABDPY_secure_det3x3(a,b,c,d,e,f,g,h,i) << ", det3x3 "
                         << signe3(a,b,c,d,e,f,g,h,i) <<  ", AR_det3 "
                         << AR_det3(a,b,c,d,e,f,g,h,i) << endl;
          cerr<<"a= ";exact_out(cerr,a); cerr<<endl;
          cerr<<"b= ";exact_out(cerr,b); cerr<<endl;
          cerr<<"c= ";exact_out(cerr,c); cerr<<endl;
          cerr<<"d= ";exact_out(cerr,d); cerr<<endl;
          cerr<<"e= ";exact_out(cerr,e); cerr<<endl;
          cerr<<"f= ";exact_out(cerr,f); cerr<<endl;
          cerr<<"g= ";exact_out(cerr,g); cerr<<endl;
          cerr<<"h= ";exact_out(cerr,h); cerr<<endl;
          cerr<<"i= ";exact_out(cerr,i); cerr<<endl;
                  cerr<<endl<<"send the above lines to spion@sophia.inria.fr"<<endl;
        }
                if (!argc)
                {
                        t1 = clock();
                        for (j=0; j<N; j++)
                                (void) FF3(a,b,c,d,e,f,g,h,i);
                        t2 = clock()-t1;
                }
                time[method] += t2 / (double) N;
          } // end for method
        } //end for trial
        if(!argc)for ( method = 0 ; method < 7 ; method++)  cout <<"&" <<  time[method]/n;
        if(!argc)cout << "\\\\\\hline";
    cout << endl;
  }  // end for input


  if (!argc){
  cout <<"\\end{tabular}" << endl;
  cout <<"\\end{document}" << endl;
  }

}
